#ifdef _WIN32
#  include <Windows.h>
#  define NOTIFY(msg) MessageBox(NULL, TEXT(msg), TEXT("fBAfUCZ~i[P"), MB_ICONERROR | MB_OK)
#else
#  define NOTIFY(msg) std::cerr << msg << '\n'
#endif
#include <cassert>

// EBhE֘Ȁ
#include "Window.h"

// bV
#include "Mesh.h"

// ʂ̊֐
#include "common.h"

// OpenCV
#include <opencv2/opencv.hpp>

namespace
{
  // 
  const GgSimpleShader::Light light =
  {
    { 0.2f, 0.2f, 0.2f, 1.0f },                           // 
    { 1.0f, 1.0f, 1.0f, 0.0f },                           // gUˌ
    { 1.0f, 1.0f, 1.0f, 0.0f },                           // ʌ
    { 0.0f, 0.0f, 1.0f, 0.0f }                            // ʒu
  };

  // ގ
  const GgSimpleShader::Material material =
  {
    { 0.8f, 0.8f, 0.8f, 1.0f },                           // ̔ˌW
    { 0.8f, 0.8f, 0.8f, 1.0f },                           // gUˌW
    { 0.2f, 0.2f, 0.2f, 1.0f },                           // ʔˌW
    50.0f                                                 // PW
  };

  // f[^
  struct Data
  {
    // _f[^̃eNX`
    GLuint point;

    // Ff[^̃eNX`
    GLuint color;

    // f[^̃CX^X̃|C^
    static Data *pData;

    // RXgN^
    Data()
    {
      glGenTextures(1, &point);
      glGenTextures(1, &color);
      if (!pData) pData = this;
    }

    // fXgN^
    virtual ~Data()
    {
      glDeleteTextures(1, &point);
      glDeleteTextures(1, &color);
    }

    // _f[^̍쐬
    void createPoint(GLsizei width, GLsizei height, const GLvoid *data = nullptr) const
    {
      // _f[^̃eNX`쐬
      glBindTexture(GL_TEXTURE_2D, point);
      glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F, width, height, 0, GL_RGB, GL_FLOAT, data);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }

    // _f[^̓]
    void submitPoint(GLsizei width, GLsizei height, const GLvoid *data) const
    {
      // _f[^̃eNX`Ƀf[^]
      glBindTexture(GL_TEXTURE_2D, point);
      glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
      glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGB, GL_FLOAT, data);
    }

    // Ff[^̍쐬
    void createColor(GLsizei width, GLsizei height, GLenum format, const GLvoid *data = nullptr) const
    {
      // Ff[^̃eNX`쐬
      glBindTexture(GL_TEXTURE_2D, color);
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, format, GL_UNSIGNED_BYTE, data);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }

    // Ff[^̓]
    void submitColor(GLsizei width, GLsizei height, GLenum format, const GLvoid *data) const
    {
      // Ff[^̃eNX`Ƀf[^]
      glBindTexture(GL_TEXTURE_2D, color);
      glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, format, GL_UNSIGNED_BYTE, data);
    }
  };

  // f[^̃|C^
  Data *Data::pData(nullptr);

  // bV̕ƍ
  int pointSize[2];

  // cv::Mat  type()  OpenGL  format ɕϊ
  GLenum toFormat(int type)
  {
    switch (type)
    {
    case CV_8UC1:
      glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
      return GL_RED;
    case CV_8UC2:
      glPixelStorei(GL_UNPACK_ALIGNMENT, 2);
      return GL_RG;
    case CV_8UC4:
      glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
      return GL_BGRA;
    default:
      assert(false);
    case CV_8UC3:
      break;
    }

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    return GL_BGR;
  }
}

// _f[^̍쐬
void createPoint(int width, int height, const void *point)
{
  pointSize[0] = width;
  pointSize[1] = height;

  Data::pData->createPoint(width, height, point);
}

// _f[^̍쐬
void createPoint(const cv::Mat &point)
{
  assert(point.type() == CV_32FC3);
  createPoint(point.cols, point.rows, point.data);
}

// _f[^̓]
void submitPoint(int width, int height, const void *point)
{
  Data::pData->submitPoint(width, height, point);
}

// _f[^̓]
void submitPoint(const cv::Mat &point)
{
  assert(point.type() == CV_32FC3);
  submitPoint(point.cols, point.rows, point.data);
}

// Ff[^̍쐬
void createColor(int width, int height, const void *color)
{
  Data::pData->createColor(width, height, GL_BGR, color);
}

// Ff[^̍쐬
void createColor(const cv::Mat &color)
{
  Data::pData->createColor(color.cols, color.rows, toFormat(color.type()), color.data);
}

// Ff[^̓]
void submitColor(int width, int height, const void *color)
{
  Data::pData->submitColor(width, height, GL_BGR, color);
}

// Ff[^̓]
void submitColor(const cv::Mat &color)
{
  Data::pData->submitColor(color.cols, color.rows, toFormat(color.type()), color.data);
}

// ̎擾
float getTime()
{
  return static_cast<float>(glfwGetTime());
}

//
// CvO
//
int main()
{
  // EBhEJ
  Window window;

  // EBhEJǂeXg
  if (!window.get())
  {
    NOTIFY("EBhEJ܂ł");
    return EXIT_FAILURE;
  }

  // f[^쐬
  Data data;

  // ݒ
  if (!setup())
  {
    NOTIFY("Ɏs܂");
    return EXIT_FAILURE;
  }

  // bV쐬
  Mesh mesh(pointSize);

  // r[ϊs߂
  const GgMatrix mv(ggLookat(0.0f, 0.0f, 5.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f));

  // Aʏs
  glEnable(GL_DEPTH_TEST);

  // wʃ|S`悷
  glDisable(GL_CULL_FACE);

  // wiFݒ肷
  glClearColor(0.3f, 0.4f, 0.5f, 0.0f);

  // EBhE܂ŌJԂ
  while (!window.shouldClose())
  {
    // XV
    if (!update()) break;

    // ʂNA
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // VF[_Lɂ
    mesh.shader.use();

    // ƍގݒ肷
    mesh.shader.loadLight(light);
    mesh.shader.loadMaterial(material);

    // _eNX`̃TvOɎgpeNX`jbgݒ肷
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, data.point);

    // FeNX`̃TvOɎgpeNX`jbgݒ肷
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, data.color);

    // ϊsݒ肷
    mesh.shader.loadMatrix(ggPerspective(1.0f, window.getAspect(), 1.0f, 10.0f), mv * window.getLeftTrackball());

    // bV`悷
    mesh.draw();

    // _uobt@O
    window.swapBuffers();
  }
}
